import _slicedToArray from "@babel/runtime/helpers/slicedToArray";
import { pipe, compose, composeRight, invoke, values, lets, guard, condS, otherwise, eq, guardV, die, tryCatch, ifNil, findIndex, reduce, appendM, tap } from 'stick-js/es';
import { foldMockSearchInit, foldMockSearch, foldMockGenerateSecret, foldMockAuthenticate } from './types.mjs';
import { discardStackTrace, ifEmptyList, randInt } from './util.mjs';

/* Takes the dictionary of chances, e.g.
 *     { case1: 10, case2: 2, case3: 5, }
 * and returns a list whose values we can compare with a random int, and the
 * last value of the list:
 *     [[10, 12, 17], 17]
 */

var makeDistribution = function makeDistribution(o) {
  return pipe(pipe(o, values), reduce(function (_ref, y) {
    var _ref2 = _slicedToArray(_ref, 2),
      xs = _ref2[0],
      n = _ref2[1];
    return [pipe(xs, appendM(n + y)), n + y];
  }, [[], 0]));
};
var mockThing = function mockThing(which) {
  return function (mockChances) {
    return invoke(function () {
      var _makeDistribution = makeDistribution(mockChances[which]),
        _makeDistribution2 = _slicedToArray(_makeDistribution, 2),
        theChances = _makeDistribution2[0],
        n = _makeDistribution2[1];
      var f = function f(y) {
        return pipe(theChances, findIndex(function (x) {
          return y < x;
        }));
      };
      return function () {
        for (var _len = arguments.length, fs = new Array(_len), _key = 0; _key < _len; _key++) {
          fs[_key] = arguments[_key];
        }
        return lets(function () {
          return randInt(n);
        }, function (y) {
          return fs[f(y)]();
        });
      };
    });
  };
};
var mockSearchInit = mockThing('searchInit');
var mockSearch = mockThing('search');
var mockGenerateSecret = mockThing('generateSecret');
var mockAuthenticate = mockThing('authenticate');
export var searchInit = function searchInit(drup, mock) {
  return function (ctrlNum, busNum, authNum, owConfig) {
    return tryCatch(function (presence) {
      return pipe(presence, ifNil(function () {
        return [null, null, false];
      }, function () {
        return pipe(presence, condS([pipe(eq('presence-with-short'), guardV([null, true, false])), pipe(eq('presence-ready'), guardV([null, true, true])), pipe(otherwise, guard(function () {
          return guardV(['Unexpected value for presence: ' + presence, null, null]);
        }))]));
      }));
    }, function (e) {
      return ['Error on searchInit (): ' + discardStackTrace(e), null, false];
    }, function () {
      return pipe(mock, foldMockSearchInit(
      // --- no mock
      function () {
        return drup.searchInit(ctrlNum, busNum, authNum, owConfig);
      },
      // --- random or specific
      function (mockChances, arg) {
        return mockSearchInit(mockChances)(function () {
          return die('[mock] I/O error during searchInit:', arg);
        }, function () {
          return null;
        }, function () {
          return 'presence-with-short';
        }, function () {
          return 'presence-ready';
        });
      }));
    });
  };
};
export var search = function search(drup, mock) {
  return function (ctrlNum, authNum) {
    return tryCatch(ifEmptyList(function () {
      return [null, [], true];
    }, function (results) {
      return [null, results, false];
    }), function (e) {
      return ['Error on search (): ' + discardStackTrace(e), null, false];
    }, function () {
      return pipe(mock, foldMockSearch(
      // --- no mock
      function () {
        return drup.search(ctrlNum, authNum);
      },
      // --- random or specific
      function (mockChances, arg) {
        return mockSearch(mockChances)(function () {
          return die('[mock] I/O error during search:', arg);
        }, function () {
          return [];
        }, function () {
          return lets(function () {
            return arg;
          }, function (deviceId) {
            return [deviceId];
          });
        });
      },
      // --- other
      function () {
        return [];
      }));
    });
  };
};
export var generateSecret = function generateSecret(drup, mock) {
  return function (ctrlNum, authNum, deviceId) {
    return tryCatch(function (secret) {
      return [null, secret];
    }, function (e) {
      return [e, null];
    }, function () {
      return pipe(mock, foldMockGenerateSecret(
      // --- no mock
      function () {
        return drup.generateSecret(ctrlNum, authNum, deviceId);
      },
      // --- random or specific
      function (mockChances, arg) {
        return mockGenerateSecret(mockChances)(function () {
          return die('[mock] I/O error during generateSecret:', arg);
        }, function () {
          return lets(function () {
            return arg;
          }, function (secret) {
            return secret;
          });
        });
      },
      // --- other
      function () {}));
    });
  };
};
export var authenticate = function authenticate(drup, mock) {
  return function (ctrlNum, authNum, deviceId, secret) {
    return tryCatch(function (authenticated) {
      return [null, authenticated];
    }, function (e) {
      return [e, null];
    }, function () {
      return pipe(mock, foldMockAuthenticate(
      // --- no mock
      function () {
        return drup.authenticate(ctrlNum, authNum, deviceId, secret);
      },
      // --- random
      function (mockChances, arg) {
        return mockAuthenticate(mockChances)(function () {
          return die('[mock] I/O error during authenticate');
        }, function () {
          return lets(function () {
            return arg;
          }, function (granted) {
            return granted;
          });
        });
      },
      // --- other
      function () {}));
    });
  };
};