import _toConsumableArray from "@babel/runtime/helpers/toConsumableArray";
import _toArray from "@babel/runtime/helpers/toArray";
import _slicedToArray from "@babel/runtime/helpers/slicedToArray";
import { pipe, compose, composeRight, filter, invoke, T, F, noop, each, sprintfN, id, prop, map, lets, not, die, tryCatch, ifNil } from 'stick-js/es';
import { createRequire } from 'module';
var require = createRequire(import.meta.url);
var druppelMod = require('../../build/Release/w1druppel');
import { searchInit, search, authenticate, generateSecret } from './drup.mjs';
import { toMock } from './types.mjs';
import { discardStackTrace, flatMap, flatMapX, flatRepeatF, isEmptyList, N } from './util.mjs';

/*
 * If the user's error handler throws an exception, we catch it, but we
 * don't do anything special after that. In the case of the 'should'
 * functions we return the obvious conservative value (false, null) etc. and
 * for search we return an empty list.
 *
 * If the user requires something special to happen when their own handler
 * throws an exception, then they should use put a try/catch around it.
 */

var tryHandler = invoke(function () {
  var error = function error(tag, e) {
    return console.error(pipe([tag, pipe(e, discardStackTrace)], sprintfN('tryHandler (): caught error during event handler %s (): %s')));
  };
  // --- `onOk/Error` means, if the handler itself throws an error.
  return function (tag, onOk, onError, handler) {
    return tryCatch(onOk, function (e) {
      return error(tag, e), onError(e);
    }, handler);
  };
});

/* Interact means generate secret or authenticate.
 * `shouldFunctionDone` returns a (possibly empty) list of device ids (for
 * `shouldGenerateSecret` or a (possibly empty) list of tuples (deviceId,
 * secret) (for `shouldAuthenticate`)
 */
var interact = function interact(tag1, tag2, shouldFunction, shouldFunctionDone, doFunction, onComplete) {
  return function (results) {
    pipe(pipe(results, flatMap(function (deviceId) {
      // --- `shouldOrSecret` is 'should' in the case of `shouldAuthenticate`
      // and 'secret' in the case of `shouldGenerateSecret`
      var _tryHandler = tryHandler(tag1, function (shouldOrSecret) {
          return [null, shouldOrSecret];
        }, function (e) {
          return [e, null];
        }, function () {
          return shouldFunction(deviceId);
        }),
        _tryHandler2 = _slicedToArray(_tryHandler, 2),
        err = _tryHandler2[0],
        shouldOrSecret = _tryHandler2[1];
      return pipe(err, ifNil(
      // --- should is the secret in the case of authenticate
      function () {
        return shouldFunctionDone(shouldOrSecret, deviceId);
      }, function (e) {
        console.error('Error during handler', tag1 + ':', discardStackTrace(e));
        return [];
      }));
    })), each(function (_ref) {
      var _ref2 = _toArray(_ref),
        deviceId = _ref2[0],
        rest = _ref2.slice(1);
      // --- result is secret / authenticateSucceeds
      var _doFunction = doFunction.apply(void 0, [deviceId].concat(_toConsumableArray(rest))),
        _doFunction2 = _slicedToArray(_doFunction, 2),
        err = _doFunction2[0],
        result = _doFunction2[1];
      return pipe(err, ifNil(function () {
        return tryHandler(tag2, noop, noop, function () {
          return onComplete(deviceId, result);
        });
      }, function (e) {
        console.error('Error during handler', tag2 + ':', discardStackTrace(e));
        return [];
      }));
    }));
  };
};

// --- returns an empty list in all cases (see `once`))

var inner = function inner(drup, mock) {
  return function (ctrlNum, ctrlTag, busNum, authNum, owConfig) {
    return function (_ref3) {
      var _ref3$onError = _ref3.onError,
        onError = _ref3$onError === void 0 ? noop : _ref3$onError,
        _ref3$onShort = _ref3.onShort,
        onShort = _ref3$onShort === void 0 ? noop : _ref3$onShort,
        _ref3$onPresence = _ref3.onPresence,
        onPresence = _ref3$onPresence === void 0 ? T : _ref3$onPresence,
        _ref3$onResult = _ref3.onResult,
        onResult = _ref3$onResult === void 0 ? T : _ref3$onResult,
        _ref3$shouldGenerateS = _ref3.shouldGenerateSecret,
        shouldGenerateSecret = _ref3$shouldGenerateS === void 0 ? F : _ref3$shouldGenerateS,
        _ref3$onGenerateSecre = _ref3.onGenerateSecret,
        onGenerateSecret = _ref3$onGenerateSecre === void 0 ? noop : _ref3$onGenerateSecre,
        _ref3$shouldAuthentic = _ref3.shouldAuthenticate,
        shouldAuthenticate = _ref3$shouldAuthentic === void 0 ? N : _ref3$shouldAuthentic,
        _ref3$onAuthenticate = _ref3.onAuthenticate,
        onAuthenticate = _ref3$onAuthenticate === void 0 ? noop : _ref3$onAuthenticate;
      var info = {
        ctrlNum: ctrlNum,
        ctrlTag: ctrlTag,
        busNum: busNum,
        authNum: authNum
      };
      var _searchInit = searchInit(drup, mock)(ctrlNum, busNum, authNum, owConfig),
        _searchInit2 = _slicedToArray(_searchInit, 3),
        errorInit = _searchInit2[0],
        presence = _searchInit2[1],
        ready = _searchInit2[2];
      if (errorInit) return tryHandler('onError', function () {
        return [];
      }, function () {
        return [];
      }, function () {
        return onError(errorInit, info);
      });
      if (!presence) return [];
      if (not(ready)) return tryHandler('onShort', function () {
        return [];
      }, function () {
        return [];
      }, function () {
        return onShort(info);
      });

      // --- @future collapse shorts perhaps, meaning if you see a short on one
      // authenticator then maybe it's not useful to generate another
      // notification for subsequent ones

      if (!tryHandler('onPresence', id, F, function () {
        return onPresence(info);
      })) return [];
      var _search = search(drup, mock)(ctrlNum, authNum),
        _search2 = _slicedToArray(_search, 3),
        errorSearch = _search2[0],
        results = _search2[1],
        empty = _search2[2];
      if (errorSearch) return tryHandler('onError', function () {
        return [];
      }, function () {
        return [];
      }, function () {
        return onError(errorSearch, info);
      });
      if (empty) return [];

      // --- `onResult` returning false means drop this result from further
      // processing (either an error occurred during the handler, or the user's
      // `onResult` callback indicated that they don't want to continue)
      var resultsRemaining = pipe(results, filter(function (deviceId) {
        return tryHandler('onResult', id, F, function () {
          return onResult(deviceId, info);
        });
      }));
      if (pipe(resultsRemaining, isEmptyList)) return [];
      pipe(resultsRemaining, interact('shouldGenerateSecret', 'onGenerateSecret', function (deviceId) {
        return shouldGenerateSecret(deviceId, info);
      }, function (should, deviceId, _) {
        return should ? [[deviceId]] : [];
      }, function (deviceId) {
        return generateSecret(drup, mock)(ctrlNum, authNum, deviceId);
      }, function (deviceId, secret) {
        return onGenerateSecret(deviceId, secret, info);
      }));
      pipe(resultsRemaining, interact('shouldAuthenticate', 'onAuthenticate', function (deviceId) {
        return shouldAuthenticate(deviceId, info);
      }, function (secret, deviceId) {
        return secret ? [[deviceId, secret]] : [];
      }, function (deviceId, secret) {
        return authenticate(drup, mock)(ctrlNum, authNum, deviceId, secret);
      }, function (deviceId, succeeds) {
        return onAuthenticate(deviceId, succeeds, info);
      }));
      return [];
    };
  };
};

/* Returns an empty list in all cases, success or failure. Maybe we'll
 * think of something useful to put in this array at some point, but since
 * everything is handled by handlers, that time is not now.
 *
 * @future Add a busNumPredicate function to only search certain buses
 * perhaps.
 */
var _once = function once(drup, mock) {
  return function (ctrlTags, authenticators, owConfig) {
    return function (handlers) {
      return pipe(ctrlTags, flatMapX(function (ctrlTag, ctrlNum) {
        return lets(
        // @throws
        function () {
          return drup.getNumBuses(ctrlNum);
        }, function (numBuses) {
          return pipe(numBuses, flatRepeatF(function (busNum) {
            return pipe(authenticators, flatMapX(function (_auth, authNum) {
              return inner(drup, mock)(ctrlNum, ctrlTag, busNum, authNum, owConfig)(handlers);
            }));
          }));
        });
      }));
    };
  };
};

// --- `mock`: `false`, 'random-normal', 'random-fast', 'random-fastest', ...
export var init = function init(_ref4) {
  var controllers = _ref4.controllers,
    authenticators = _ref4.authenticators,
    _ref4$verbose = _ref4.verbose,
    verbose = _ref4$verbose === void 0 ? false : _ref4$verbose;
  var _ref5 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
    _ref5$mock = _ref5.mock,
    mock = _ref5$mock === void 0 ? false : _ref5$mock;
  var _ref6 = pipe(controllers, ifNil(function () {
      return die('missing key “controllers”');
    }, function (cs) {
      return [// --- mapping from ctrl num to tag
      pipe(cs, map(prop('tag'))), pipe(cs, map(prop('config')))];
    })),
    _ref7 = _slicedToArray(_ref6, 2),
    ctrlTags = _ref7[0],
    ctrls = _ref7[1];
  var config = {
    verbose: verbose,
    authenticators: authenticators,
    controllers: ctrls
  };

  // --- also validates
  var drup = druppelMod.init(config, {
    mock: Boolean(mock)
  });
  return {
    once: function once(owConfig, handlers) {
      var mockSpec = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      return _once(drup, toMock(mockSpec))(ctrlTags, authenticators, owConfig)(handlers);
    },
    forever: function forever(owConfig, ms, handlers) {
      var _this = this;
      var mockSpec = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : false;
      var handle = {
        _cancel: false,
        cancel: function cancel() {
          this._cancel = true;
        }
      };
      var loop = function loop() {
        _this.once(owConfig, handlers, mockSpec);
        if (handle._cancel) return;
        setTimeout(loop, ms);
      };
      setTimeout(loop, 0);
      return handle;
    },
    cleanup: function cleanup() {
      drup.cleanup();
    }
  };
};