import _defineProperty from "@babel/runtime/helpers/defineProperty";
import { pipe, compose, composeRight, invoke, ifOk, id, mergeTo } from 'stick-js/es';

// --- must not depend on types
import { base64encodeFromU8Array } from './util.mjs';
export var mockDeviceId = '0123456789abcdef';
export var mockError = 'mock-failure';
export var mockSecret = new Uint8Array([0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7]);
export var mockSecretBase64 = base64encodeFromU8Array(mockSecret);

// --- total per block is arbitrary (these are relative weights)
export var getMockChances = invoke(function () {
  var o = {
    'random-normal': function randomNormal() {
      return {
        searchInit: {
          error: 1,
          noPresence: 7000,
          presenceWithShort: 15,
          presenceReady: 15
        },
        search: {
          error: 1,
          zeroResults: 100,
          oneResult: 500
        },
        generateSecret: {
          error: 1,
          ok: 100
        },
        authenticate: {
          error: 1,
          authenticate: 10
        }
      };
    },
    'random-fast': function randomFast() {
      return {
        searchInit: {
          error: 1,
          noPresence: 700,
          presenceWithShort: 15,
          presenceReady: 15
        },
        search: {
          error: 1,
          zeroResults: 100,
          oneResult: 500
        },
        generateSecret: {
          error: 1,
          ok: 100
        },
        authenticate: {
          error: 1,
          authenticate: 10
        }
      };
    },
    'random-fastest': function randomFastest() {
      return {
        searchInit: {
          error: 1,
          noPresence: 100,
          presenceWithShort: 15,
          presenceReady: 15
        },
        search: {
          error: 1,
          zeroResults: 100,
          oneResult: 500
        },
        generateSecret: {
          error: 1,
          ok: 100
        },
        authenticate: {
          error: 1,
          authenticate: 10
        }
      };
    },
    searchInit: function searchInit(which) {
      return {
        searchInit: pipe(_defineProperty({}, which, 1), mergeTo({
          error: 0,
          noPresence: 0,
          presenceWithShort: 0,
          presenceReady: 0
        }))
      };
    },
    search: function search(which) {
      return {
        search: pipe(_defineProperty({}, which, 1), mergeTo({
          error: 0,
          zeroResults: 0,
          oneResult: 0
        }))
      };
    },
    generateSecret: function generateSecret(which) {
      return {
        generateSecret: pipe(_defineProperty({}, which, 1), mergeTo({
          error: 0,
          ok: 0
        }))
      };
    },
    authenticate: function authenticate(which) {
      return {
        authenticate: pipe(_defineProperty({}, which, 1), mergeTo({
          error: 0,
          authenticate: 0
        }))
      };
    }
  };
  return function (profile) {
    if (profile === false) return null;
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }
    return pipe(o[profile].apply(o, args), ifOk(id, function () {
      console.error('getMockChances (): invalid profile ' + profile + ', using default');
      return o.get(true);
    }));
  };
});