import { pipe, compose, composeRight, T, F, invoke, die } from 'stick-js/es';
import daggy from 'daggy';
import { cata } from 'alleycat-js/es/bilby';
import { getMockChances, mockDeviceId, mockError, mockSecret } from './mock.mjs';
import { randBool } from './util.mjs';
var Mock = daggy.taggedSum('Mock', {
  MockNone: [],
  MockRandom: ['chances'],
  MockSearchInitError: ['msg'],
  MockSearchInitNoPresence: [],
  MockSearchInitPresenceWithShort: [],
  MockSearchInitPresenceReady: [],
  MockSearchError: ['msg'],
  MockSearchResults: ['deviceIds'],
  MockGenerateSecretError: ['msg'],
  MockGenerateSecretOk: ['deviceId', 'secret'],
  MockAuthenticateError: ['msg'],
  MockAuthenticateAuthenticate: ['succeeds'],
  MockTestProbe: ['deviceId', 'succeeds']
});
var MockNone = Mock.MockNone,
  MockRandom = Mock.MockRandom,
  MockSearchInitError = Mock.MockSearchInitError,
  MockSearchInitNoPresence = Mock.MockSearchInitNoPresence,
  MockSearchInitPresenceWithShort = Mock.MockSearchInitPresenceWithShort,
  MockSearchInitPresenceReady = Mock.MockSearchInitPresenceReady,
  MockSearchError = Mock.MockSearchError,
  MockSearchResults = Mock.MockSearchResults,
  MockGenerateSecretError = Mock.MockGenerateSecretError,
  MockGenerateSecretOk = Mock.MockGenerateSecretOk,
  MockAuthenticateError = Mock.MockAuthenticateError,
  MockAuthenticateAuthenticate = Mock.MockAuthenticateAuthenticate,
  MockTestProbe = Mock.MockTestProbe;
export { MockNone, MockRandom, MockSearchInitError, MockSearchInitNoPresence, MockSearchInitPresenceWithShort, MockSearchInitPresenceReady, MockSearchError, MockSearchResults, MockGenerateSecretError, MockGenerateSecretOk, MockAuthenticateError, MockAuthenticateAuthenticate, MockTestProbe };
export var toMock = invoke(function () {
  var m = new Map([[false, MockNone], ['random-normal', MockRandom(getMockChances('random-normal'))], ['random-fast', MockRandom(getMockChances('random-fast'))], ['random-fastest', MockRandom(getMockChances('random-fastest'))], ['search-init-error', MockSearchInitError(mockError)], ['search-init-no-presence', MockSearchInitNoPresence], ['search-init-presence-with-short', MockSearchInitPresenceWithShort], ['search-init-presence-ready', MockSearchInitPresenceReady], ['search-error', MockSearchError(mockError)], ['search-zero-results', MockSearchResults([])], ['search-one-result', MockSearchResults([mockDeviceId])], ['generate-secret-error', MockGenerateSecretError(mockError)], ['generate-secret-ok', MockGenerateSecretOk(mockDeviceId, mockSecret)], ['authenticate-error', MockAuthenticateError(mockError)],
  // ------ fails and succeeds refer here to the `authenticate` step,
  // where succeeds means that the secret matches the secret in the device
  // and there are no I/O errors.
  ['authenticate-fails', MockAuthenticateAuthenticate(false)], ['authenticate-succeeds', MockAuthenticateAuthenticate(true)],
  /* These are the ones to use for testing kattenluik as they most closely
   * mimic a key being held in front of the probe. The flow will then
   * continue with (mock) generate secret and (mock) authenticate
   * according to the caller's own `shouldGenerateSecret` and
   * `shouldAuthenticate` logic, which is why this one is useful. It is
   * implemented by setting the probabilities for searchInit, search,
   * generateSecret, and authenticate all to 1.
   */
  ['test-probe-authenticate-fails', MockTestProbe(mockDeviceId, false)], ['test-probe-authenticate-succeeds', MockTestProbe(mockDeviceId, true)]]);
  return function (mockSpec) {
    var _m$get;
    return (_m$get = m.get(mockSpec)) !== null && _m$get !== void 0 ? _m$get : die('Invalid mock value: ' + String(mockSpec));
  };
});
export var mockBool = cata({
  MockNone: F,
  MockRandom: T,
  MockSearchInitError: T,
  MockSearchInitNoPresence: T,
  MockSearchInitPresenceWithShort: T,
  MockSearchInitPresenceReady: T,
  MockSearchError: T,
  MockSearchResults: T,
  MockGenerateSecretError: T,
  MockGenerateSecretOk: T,
  MockAuthenticateError: T,
  MockAuthenticateAuthenticate: T,
  MockTestProbe: T
});

/* We provide two kinds of mocking: 'random', in which events are fired
 * randomly according to the probabilities given by `getMockChances`, and
 * 'specific', in which we can cause a specific event to fire (e.g.
 * 'generate-secret-error' or 'authenticate-deny'). They both route through
 * the same function, which is why it's called `randomOrSpecific`. In the
 * specific case we alter the probability distribution so that the case we
 * want is 1 and all the others are 0.
 *
 * Note that for the specific case, when mocking for example a generate
 * secret event, we also need to make sure we mock searchInit and search
 * with 100% probability, or else we will never reach the point we want.
 * This is done in the functions at the top of each fold function (e.g.
 * `ready`, `gotResults`, etc.).
 */

export var foldMockSearchInit = invoke(function () {
  // @todo mockDeviceId ('arg') is not used here
  var ready = function ready(randomOrSpecific) {
    return randomOrSpecific(getMockChances('searchInit', 'presenceReady'), mockDeviceId);
  };
  return function (none, randomOrSpecific) {
    return cata({
      MockNone: function MockNone() {
        return none();
      },
      MockRandom: function MockRandom(chances) {
        return randomOrSpecific(chances);
      },
      MockSearchInitError: function MockSearchInitError(msg) {
        return randomOrSpecific(getMockChances('searchInit', 'error'), msg);
      },
      MockSearchInitNoPresence: function MockSearchInitNoPresence() {
        return randomOrSpecific(getMockChances('searchInit', 'noPresence'));
      },
      MockSearchInitPresenceWithShort: function MockSearchInitPresenceWithShort() {
        return randomOrSpecific(getMockChances('searchInit', 'presenceWithShort'));
      },
      MockSearchInitPresenceReady: function MockSearchInitPresenceReady() {
        return ready(randomOrSpecific);
      },
      MockSearchError: function MockSearchError(_) {
        return ready(randomOrSpecific);
      },
      MockSearchResults: function MockSearchResults(_deviceIds) {
        return ready(randomOrSpecific);
      },
      MockGenerateSecretError: function MockGenerateSecretError(_) {
        return ready(randomOrSpecific);
      },
      MockGenerateSecretOk: function MockGenerateSecretOk(_) {
        return ready(randomOrSpecific);
      },
      MockAuthenticateError: function MockAuthenticateError(_) {
        return ready(randomOrSpecific);
      },
      MockAuthenticateAuthenticate: function MockAuthenticateAuthenticate() {
        return ready(randomOrSpecific);
      },
      MockTestProbe: function MockTestProbe() {
        return ready(randomOrSpecific);
      }
    });
  };
});
export var foldMockSearch = invoke(function () {
  var zeroOrOneResult = function zeroOrOneResult(randomOrSpecific) {
    var haveResult = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
    return randomOrSpecific(getMockChances('search', haveResult ? 'oneResult' : 'zeroResults'), mockDeviceId);
  };
  return function (none, randomOrSpecific, other) {
    return cata({
      MockNone: function MockNone() {
        return none();
      },
      MockRandom: function MockRandom(chances) {
        return randomOrSpecific(chances, mockDeviceId);
      },
      MockSearchInitError: function MockSearchInitError(_) {
        return other();
      },
      MockSearchInitNoPresence: function MockSearchInitNoPresence() {
        return other();
      },
      MockSearchInitPresenceWithShort: function MockSearchInitPresenceWithShort() {
        return other();
      },
      MockSearchInitPresenceReady: function MockSearchInitPresenceReady() {
        return other();
      },
      MockSearchError: function MockSearchError(msg) {
        return randomOrSpecific(getMockChances('search', 'error'), msg);
      },
      MockSearchResults: function MockSearchResults(_deviceIds) {
        var have = _deviceIds.length !== 0;
        return zeroOrOneResult(randomOrSpecific, have);
      },
      MockGenerateSecretError: function MockGenerateSecretError(_) {
        return zeroOrOneResult(randomOrSpecific);
      },
      MockGenerateSecretOk: function MockGenerateSecretOk(_) {
        return zeroOrOneResult(randomOrSpecific);
      },
      MockAuthenticateError: function MockAuthenticateError(_) {
        return zeroOrOneResult(randomOrSpecific);
      },
      MockAuthenticateAuthenticate: function MockAuthenticateAuthenticate() {
        return zeroOrOneResult(randomOrSpecific);
      },
      MockTestProbe: function MockTestProbe() {
        return zeroOrOneResult(randomOrSpecific);
      }
    });
  };
});
export var foldMockGenerateSecret = invoke(function () {
  var generatedSecret = function generatedSecret(randomOrSpecific) {
    return randomOrSpecific(getMockChances('generateSecret', 'ok'), mockSecret);
  };
  return function (none, randomOrSpecific, other) {
    return cata({
      MockNone: function MockNone() {
        return none();
      },
      MockRandom: function MockRandom(chances) {
        return randomOrSpecific(chances, mockSecret);
      },
      MockSearchInitError: function MockSearchInitError(_) {
        return other();
      },
      MockSearchInitNoPresence: function MockSearchInitNoPresence() {
        return other();
      },
      MockSearchInitPresenceWithShort: function MockSearchInitPresenceWithShort() {
        return other();
      },
      MockSearchInitPresenceReady: function MockSearchInitPresenceReady() {
        return other();
      },
      MockSearchError: function MockSearchError(_) {
        return other();
      },
      MockSearchResults: function MockSearchResults(_) {
        return other();
      },
      MockGenerateSecretError: function MockGenerateSecretError(_) {
        return randomOrSpecific(getMockChances('generateSecret', 'error'), mockError);
      },
      MockGenerateSecretOk: function MockGenerateSecretOk(_deviceId, _secret) {
        return generatedSecret(randomOrSpecific);
      },
      MockAuthenticateError: function MockAuthenticateError(_) {
        return generatedSecret(randomOrSpecific);
      },
      MockAuthenticateAuthenticate: function MockAuthenticateAuthenticate() {
        return generatedSecret(randomOrSpecific);
      },
      MockTestProbe: function MockTestProbe() {
        return generatedSecret(randomOrSpecific);
      }
    });
  };
});
export var foldMockAuthenticate = function foldMockAuthenticate(none, randomOrSpecific, other) {
  return invoke(function () {
    var authenticate = function authenticate(randomOrSpecific, succeeds) {
      return randomOrSpecific(getMockChances('authenticate', 'authenticate'), succeeds);
    };
    return cata({
      MockNone: function MockNone() {
        return none();
      },
      MockRandom: function MockRandom(chances) {
        return randomOrSpecific(chances, randBool());
      },
      MockSearchInitError: function MockSearchInitError(_) {
        return other();
      },
      MockSearchInitNoPresence: function MockSearchInitNoPresence() {
        return other();
      },
      MockSearchInitPresenceWithShort: function MockSearchInitPresenceWithShort() {
        return other();
      },
      MockSearchInitPresenceReady: function MockSearchInitPresenceReady() {
        return other();
      },
      MockSearchError: function MockSearchError(_) {
        return other();
      },
      MockSearchResults: function MockSearchResults(_) {
        return other();
      },
      MockGenerateSecretError: function MockGenerateSecretError(_) {
        return other();
      },
      MockGenerateSecretOk: function MockGenerateSecretOk(_) {
        return other();
      },
      MockAuthenticateError: function MockAuthenticateError(_) {
        return randomOrSpecific(getMockChances('authenticate', 'error'), mockError);
      },
      MockAuthenticateAuthenticate: function MockAuthenticateAuthenticate(succeeds) {
        return authenticate(randomOrSpecific, succeeds);
      },
      MockTestProbe: function MockTestProbe(_deviceId, succeeds) {
        return authenticate(randomOrSpecific, succeeds);
      }
    });
  });
};